#!/usr/bin/python
# -*- coding: latin-1 -*-
#
#    Copyright © 2013 Kevin Cheung
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
from pymprog import *
from music21 import *
from fractions import *

# penalty factors
PENALTY_3RD = 1.0
PENALTY_P4_P5 = 5.0 
PENALTY_P1 = 10.0

# Predefined temperaments for use with score_to_csound_with_tuning
# All values specified in cents.
# Equal temperament
ET = [0.0, 100.0, 200.0, 300.0, 400.0, 500.0, \
      600.0, 700.0, 800.0, 900.0, 1000.0, 1100.0]

# Bach/Lehman
BL = [0.0, 98.04, 196.06, 298.04, 392.18, 510.96, \
      596.06, 698.04, 798.04, 894.13, 998.04, 1094.13]

# Kirnberger 3
K3 = [0.0, 90.22, 193.16, 294.13, 386.31, 498.04, \
      590.22, 696.58, 792.18, 889.74, 996.09, 1092.18]

# 1/4 Meantone Comma
QM = [0.0, 76.05, 193.16, 310.26, 386.31, 503.42, \
      579.47, 696.58, 772.63, 889.74, 1006.84, 1096.09]

# 1/6 Pythagorean Comma
SP = [0.0, 86.31, 196.09, 305.87, 392.18, 501.96, \
      588.27, 698.04, 784.36, 894.13, 1003.91, 1090.22]

# Werckmeister III
W3 = [0.0, 90.22, 192.18, 294.13, 390.22, 498.04, \
      588.27, 696.09, 792.18, 888.27, 996.09, 1092.18]

# Vallotti
VA = [0.0, 94.13, 196.09, 298.04, 392.18, 501.96, \
      592.18, 698.04, 796.09, 894.13, 1000.0, 1090.22]

# Young 2
Y2 = [0.0, 90.22, 196.09, 294.13, 392.18, 498.04, \
      588.27, 698.04, 792.18, 894.13, 996.09, 1090.22]

PER_4TH = [ ('C', 'F'), ('C#', 'F#'), ('D', 'G'), ('E-', 'A-'), \
            ('E', 'A'), ('F', 'B-'), ('F#', 'B'), ('G', 'C'), ('G#', 'C#'), \
            ('A', 'D'), ('B-', 'E-'), ('B', 'E') \
         # ignored intervals - uncomment to include
         # , ('D-', 'G-'), ('D#', 'G#'), ('E#', 'A#'), ('G-', 'C-'), \
         # ('A-', 'D-'), ('A#', 'D#'), ('B#', 'E#')
          ]

MAJ_3RD = [ ('C', 'E'), ('D-', 'F'), ('D', 'F#'), ('E-', 'G'), ('E', 'G#'), \
            ('F', 'A'), ('G', 'B'), ('A-', 'C'), ('A', 'C#'), ('B-', 'D'), \
            ('B', 'D#') \
         # ignored intervals - uncomment to include
         # ,('C#', 'E#'), ('F-', 'A-'), ('F#', 'A#'), ('G-', 'B-'), ('G#', 'B#')
          ]

MIN_3RD = [ ('C', 'E-'), ('C#', 'E'), ('D', 'F'), ('E', 'G'), ('F', 'A-'), \
            ('F#', 'A'), ('G', 'B-'), ('G#', 'B'), ('A', 'C'), ('B', 'D') \
          # unsupported intervals - uncomment to include
          # , ('D-', 'F-'), ('D#', 'F#'), ('E-', 'G-'), ('E#', 'G#'), 
          # ('A-', 'C-'), ('A#', 'C#'), ('B-', 'D-')
          ]

class MyRatio:
    def __init__(self, numerator, denominator, weight=0.0): 
        # positive weight means NON preferred
        self.fraction = Fraction(numerator, denominator)
        self.weight = weight
           
RATIOS = { 'C' : [MyRatio(1,1)], \
           # 9/8 is preferred for D 
           'D' : [MyRatio(10,9,0.1), MyRatio(9,8)], \
           'E' : [MyRatio(5,4), MyRatio(81,64)], \
           'F' : [MyRatio(4,3)], \
           'G' : [MyRatio(3,2)], \
           'A' : [MyRatio(5,3), MyRatio(27,16)], \
           # 243/128 is preferred for B
           'B' : [MyRatio(15,8,0.1), MyRatio(243,128)], \
           'F#': [MyRatio(25,18), MyRatio(45,32)], \
           'C#': [MyRatio(25,24), MyRatio(135,128)], \
           'G#': [MyRatio(25,16), MyRatio(405,256)], \
           'D#': [MyRatio(75,64), MyRatio(1215,1024)], \
           'A#': [MyRatio(125,72), MyRatio(225,128)], \
           'E#': [MyRatio(125,96), MyRatio(675,512)], \
           'B#': [MyRatio(125,64), MyRatio(2025,1024)], \
           'B-': [MyRatio(16,9), MyRatio(9,5)], \
           'E-': [MyRatio(32,27), MyRatio(6,5)], \
           'A-': [MyRatio(128,81), MyRatio(8,5)], \
           'D-': [MyRatio(256,243), MyRatio(16,15)], \
           'G-': [MyRatio(64,45), MyRatio(36,25)], \
           'C-': [MyRatio(128,135), MyRatio(24, 25)], \
           'F-': [MyRatio(512,405), MyRatio(32,25)] }

RATIOS_EXTENDED = { 
           'C' : [MyRatio(1,1)], \
           'D' : [MyRatio(10,9,0.1), MyRatio(9,8)], \
           'E' : [MyRatio(5,4), MyRatio(81,64)], \
           'F' : [MyRatio(4,3), MyRatio(27,20,0.1)], \
           'G' : [MyRatio(40,27,0.1), MyRatio(3,2), MyRatio(243,160,0.1)], \
           'A' : [MyRatio(5,3), MyRatio(27,16)], \
           'B' : [MyRatio(15,8,0.1), MyRatio(243,128), MyRatio(50,27,0.1)], \
           'F#': [MyRatio(25,18), MyRatio(45,32)], \
           'C#': [MyRatio(25,24), MyRatio(135,128)], \
           'G#': [MyRatio(25,16), MyRatio(405,256)], \
           'D#': [MyRatio(75,64), MyRatio(1215,1024)], \
           'A#': [MyRatio(125,72), MyRatio(225,128)], \
           'E#': [MyRatio(125,96), MyRatio(675,512)], \
           'B#': [MyRatio(125,64), MyRatio(2025,1024)], \
           'B-': [MyRatio(16,9), MyRatio(9,5)], \
           'E-': [MyRatio(32,27), MyRatio(6,5)], \
           'A-': [MyRatio(128,81), MyRatio(8,5)], \
           'D-': [MyRatio(256,243), MyRatio(16,15)], \
           'G-': [MyRatio(64,45), MyRatio(36,25)], \
           'C-': [MyRatio(128,135), MyRatio(24, 25)], \
           'F-': [MyRatio(512,405), MyRatio(32,25)] }

class _MyNote:
    def __init__(self, pitch, freq, dur, offset):
        self.frequency = freq;
        self.pitch = pitch;
        self.duration = dur;
        self.offset = offset;

class _NoteRatio:
    def __init__(self, note, ratioVars):
        self.note = note;
        self.ratioVars = ratioVars;

def _score_to_list(s):
    def part_to_list(p):
        l = []
        offset = 0.0
        tiedNotes = [] 
        for m in p.getElementsByClass( stream.Measure ):
            for n in m.getElementsByClass( note.Note ):
                t = _MyNote(n.pitch, n.frequency, \
                    n.duration.quarterLength, offset+n.offset)         
                if (n.tie is None):
                    l.append( t )
                else:
                    if (n.tie.type == 'start'):
                        l.append( t )
                        tiedNotes.append( t )
                    else:
                        for tiedNote in tiedNotes:
                            if (tiedNote.pitch.nameWithOctave \
                                == t.pitch.nameWithOctave):
                                tiedNote.duration += t.duration
                                if (n.tie.type == 'stop'):
                                    tiedNotes.remove( tiedNote )
                                break;
            offset += m.duration.quarterLength
        return l;
    return [ part_to_list(p) for p in s.parts ]


def _score_to_csound_work(partList, tempo, filename):
    f = open(filename, 'w')

    if (f is not None):
        # header stuff
        f.write( "<CsoundSynthesizer>\n" )
        f.write( "<CsOptions>\n" )
        f.write( "-odac\n" )
        f.write( "</CsOptions>\n" )
        f.write( "<CsInstruments>\n" )
        f.write( "sr = 44100\n" )
        f.write( "ksmps = 32\n" )
        f.write( "nchnls = 2\n" )
        f.write( "0dbfs = 1\n" )

        nParts = len(partList)

        amp = 0.95/1.41421356237/len(partList)

        for i in range(nParts):
            f.write( "instr " + str(i+1) + "\n" )
            f.write( "    adeclick linseg 0, 0.05, 1, p3-0.1, 1," + \
                     " 0.05, 0\n")
            f.write( "    asig oscili " + str(amp) + "*adeclick, p4," + \
                     str(i+1) + "\n" )
            f.write( "       out p5*asig, p6*asig\n" )
            f.write( "endin\n" )
     
        f.write( "</CsInstruments>\n" )

        f.write( "<CsScore>\n" )
        f.write( "t0 " + str(tempo) + "\n" )

        for i in range(nParts):
            # a somewhat organ-like sound for each part
            f.write( "f" + str(i+1) +" 0 4096 10 0.5 0.75 0.4 0.50 0.2 " + \
                     "0.2 0.0 0.25 0.0 0.10 0.0 0.10\n" )

        # pan the parts to obtain a reasonable soundstage
        panAngle = 0.25 * math.pi/2.0
        i = 0
        for part in partList:
            i += 1
            f.write( "\n" )
            for n in part:
                f.write( "i" + str(i) + " " + str(n.offset) + " " + \
                         str(n.duration) + \
                         " %.2f" % n.frequency)
                f.write( " %.4f %.4f" % \
                          (math.sin(panAngle), math.cos(panAngle)) + "\n")
            panAngle += 0.5 * math.pi/2.0 / (nParts - 1)
        f.write( "</CsScore>\n" )
        f.write( "</CsoundSynthesizer>\n" )
    else:
        print "Failed to open file " + filename
    

### Tune score with custom tuning ###
def score_to_csound(score, tuning=ET, tempo=80, filename="custom_tuned.csd", \
    A4=440.0):
    partList = _score_to_list(score)

    a4p = pitch.Pitch('A4').pitchClass
    octaveFreqs = [ math.pow(2.0, 
                    (tuning[i] - tuning[a4p])/1200.0) * A4 \
                    for i in range(12) ]

    for i in range(12):
       print str(octaveFreqs[i])

    # process frequencies
    for part in partList:
        for n in part:
            freq = octaveFreqs[n.pitch.pitchClass]
            freq *= math.pow(2.0, (n.pitch.octave - 4.0))
            n.frequency = freq

    _score_to_csound_work( partList, tempo, filename )


def _overlap(note1, note2):
    # a minimum of a 16th note to be considered overlapping

    startTime = max(note1.offset, note2.offset)
    endTime = min(note1.offset + note1.duration.quarterLength, \
                  note2.offset + note2.duration.quarterLength)
    return max(0.0, endTime - startTime)


def _just_ratio(pitchName1, pitchName2):
    result = None;

    if (pitchName1 == pitchName2):
        result = ( Fraction(1,1), PENALTY_P1 ) # very heavy penalty

    elif ( (pitchName1, pitchName2) in PER_4TH):
        result = ( Fraction(4,3), PENALTY_P4_P5 ) # heavy penalty

    elif ( (pitchName2, pitchName1) in PER_4TH):
        result = ( Fraction(3,2), PENALTY_P4_P5 ) # heavy penalty

    elif ( (pitchName1, pitchName2) in MIN_3RD):
        result = ( Fraction(6,5), PENALTY_3RD )

    elif ( (pitchName2, pitchName1) in MIN_3RD):
        result = ( Fraction(5,3), PENALTY_3RD )

    elif ( (pitchName1, pitchName2) in MAJ_3RD):
        result = ( Fraction(5,4), PENALTY_3RD )

    elif ( (pitchName2, pitchName1) in MAJ_3RD):
        result = ( Fraction(8,5), PENALTY_3RD )

    return result

### Tune score optimally ###
def opt(s, pitchCenter='C', tempo=80, extended=False, A4=440.0):

    frequency = 264.0 * A4/440.0 # frequency for C4

    if (extended):
        useRatios = RATIOS_EXTENDED
    else:
        useRatios = RATIOS

    transposeInterval = interval.Interval('P1')

    # transpose score to C as pitchCenter
    transposeInterval = interval.Interval(note.Note('C4'), \
                        note.Note(pitchCenter) )
    freqMultiplier = useRatios[pitchCenter][0].fraction

    score = s.transpose(transposeInterval.reverse())

    nMeasures = len(score.parts[0].getElementsByClass(stream.Measure))

    nParts = len(score.parts)

    beginModel('optune')

    # verbose(True)

    varList = []

    objCoefVars = []

    overlapVars = []

    tiedList = [ [] for j in range(nParts) ]

    prevNoteRatio = [ None for j in range(nParts) ]

    for i in range(0, nMeasures):

        print "Measure " + str(i+1)
        tmpList = [ [] for j in range(nParts)]

        for j in range(0, nParts):
            # go through consecutive notes
            tmpList.append([])

            localMeasure = score.parts[j].getElementsByClass(stream.Measure)[i]
            notes = localMeasure.getElementsByClass(note.Note)

            if (len(notes) == 0):
                prevNoteRatio[j] = None

            for n in notes:
                noteRatio = _NoteRatio(n, \
                        [ (r, var(kind=bool, name="M" + str(i+1) + "_p" + \
                              str(j+1) + "_" + \
                              n.transpose(transposeInterval).pitch.name + \
                              "_" + str(n.offset) + "_" + str(r.fraction))) \
                              for r in useRatios[n.pitch.name] ] )
                tmpList[j].append( noteRatio )

                st( sum(t[1] for t in noteRatio.ratioVars)==1 )

                for t in noteRatio.ratioVars:
                    if (t[0].weight > 0.0):
                        objCoefVars.append( (t[0].weight/100.0, t[1]) )

                if (n.tie is not None):
                    if (n.tie.type == 'start'):
                        tiedList[j].append( noteRatio )
                    else:
                        for tn in tiedList[j]:
                            if (tn.note.pitch.name==noteRatio.note.pitch.name):
                                if (n.tie.type == 'stop'):
                                    tiedList[j].remove( tn )
                                st(tn.ratioVars[r][1]-noteRatio.ratioVars[r][1]\
                                    == 0 for r in range(len(tn.ratioVars)))

                if (prevNoteRatio[j] is not None):
                    if (prevNoteRatio[j].note.pitch.name \
                        == noteRatio.note.pitch.name):
                        st( prevNoteRatio[j].ratioVars[r][1] \
                            - noteRatio.ratioVars[r][1] \
                            == 0 for r in range(len(noteRatio.ratioVars)))
                prevNoteRatio[j] = noteRatio

            varList += [ t for t in tmpList[j]]

        # process time-overlapping notes
        for p1 in range(0, nParts-1):
            for nr1 in tmpList[p1]:
                for p2 in range(p1+1, nParts):
                    for nr2 in tmpList[p2]:
                        overlapDur = _overlap(nr1.note, nr2.note)

                        if (overlapDur > 0.0):
                            pitch1 = nr1.note.pitch.name
                            pitch2 = nr2.note.pitch.name
                            justRatio = _just_ratio(pitch1, pitch2)

                            if (justRatio is not None):
                                overlapDur *= justRatio[1] # penalty
                                for ratioVars1 in nr1.ratioVars:
                                    for ratioVars2 in nr2.ratioVars:
                                        ratio = ratioVars2[0].fraction/ratioVars1[0].fraction
                                        if (ratio < 1):
                                            ratio *= 2
                                        if (ratio != justRatio[0]):
                            
                                            v = var(kind=bool, name="z_"+ \
                                                ratioVars1[1].name + "_" + \
                                                ratioVars2[1].name)

                                            overlapVars.append( v )
                                            st( v-ratioVars1[1]-ratioVars2[1] \
                                                >= -1 )
                                            objCoefVars.append( (overlapDur, v))

    minimize( sum( c * v for (c, v) in objCoefVars) )
    solve(int)
    print "Solver status: ", status()
    print 'Z = %g' % vobj()

    for nr in varList:
        n = nr.note
        for r in nr.ratioVars:
            v = r[1]
            if (v.primal > 0):
                tmpFreq = frequency * r[0].fraction
                tmpFreq *= pow(2.0, n.pitch.octave - 4)
                n.frequency = tmpFreq * freqMultiplier
                n.editorial.comment.text = 'hi'
                n.editorial.comment.position = "above"

                print str(v.name) + ", " + " freq = %.2f" % n.frequency

    print ""
    for v in overlapVars:
        if (v.primal > 0.0):
            print str(v.name)
    endModel()

    # Generate csound file
    if (extended):
        _score_to_csound_work( _score_to_list(score), tempo, 'optuned_ext.csd')
    else:
        _score_to_csound_work( _score_to_list(score), tempo, 'optuned.csd')

    return score
